/*-------------------------------------------------------------------------+
|  Dynamic memory package, functions (SUN version)                         |
|                                                                          |
| Date of last updating: 18 November 1991                                  |
|                                                                          |
| Copyright (C) 1991 by Kazimierz Subieta.                                 |
| This software is subject to a licence agreement.                         |
| No restrictions when used for didactic purposes.                         |
+_________________________________________________________________________*/


#include        <stdio.h>
#include        <dynmem.h>
#include        <malloc.h>
#include        <limits.h>
#include        <string.h>
#include        <messages.h>


#define b0  0x0001
#define b1  0x0002
#define b2  0x0004
#define b3  0x0008
#define b4  0x0010
#define b5  0x0020
#define b6  0x0040
#define b7  0x0080
#define b8  0x0100
#define b9  0x0200
#define b10 0x0400
#define b11 0x0800
#define b12 0x1000
#define b13 0x2000
#define b14 0x4000
#define b15 0x8000

typedef struct inventory
  {
    unsigned char STATUS;       /* Status of an object: b7 b6 b5 b4 b3 b2 b1 b0
				b7 = 0 - cancelled,     b7 = 1 - existing,
				b6 = 0 - main,          b6 = 1 - secondary,
				b4 = 0 - non-direct,    b4 = 1 - direct,
				b3 - b0 - size of direct object.
			   From b6, b4 only one can be 1 and for unfixed only*/

    unsigned char FRQ;      /* Frequency of an object */
    unsigned char FIXING;   /* Fixing of an object; 0 = unfixed */
    unsigned char RESERVE;  /* Optimization of speed: additional byte to
			       make 16 bytes description; it stores reserve. */
    union
      {
	struct /* For objects in main memory */
	  {
	    char * OBJBEG;      /* Beginning pointer of an object */
	    char * OBJEND;      /* Ending pointer of an object + 1;
				   (OBJEND - OBJBEG) is the object size */
	  } MAIN;
	struct  /* For objects on the secondary file */
	  {
	    long OBJBEG;        /* Beginning address for secondary objects */
	    long OBJEND;        /* Ending address + 1 for secondary objects.
				   (OBJEND - OBJBEG) is the object size */
	  } SECOND;
      } U;
    unsigned short int PRIOR;   /* Name of prior object, or 0xFFFF if no prior*/
    unsigned short int NEXT;    /* Name of next  object, or 0xFFFF if no next */
  } inventory;

/* Note: for direct objects the area from OBJBEG to NEXT (inclusively) is
   treated as a continuous area of DIRSIZE bytes, where such an object is
   stored; (b3 - b0) of STATUS store its size. For alignment to 4, in DIRECT
   objects two bytes, FIXING and RESERVE are not used */

/*=============================================== Note on garbage lists ====*/
/* Information about garbage areas is stored in two double-chained lists, one
for the main memory, and one for the secondary. Entries to these lists are
ENT_NE_MAIN (for main) and ENT_NE_SECO (for secondary). The EOL is marked by
0xFFFF. Lists contain (unsigned) names of objects. NEXT and PRIOR members of
an inventory item (inventory) contain proper names if the item is not direct.
The garbage between two neighbouring elements of item I1 and next item I2 is
determined by (I1->U.MAIN.OBJEND) - beginning of the garbage area, and
(I2->U.MAIN.OBJBEG) - a pointer to char after the garbage area;
I1, I2 are pointers to proper inventory items. */

#define EXISTING    0x80
#define SECONDARY   0x40
#define DIRECT      0x10
#define DIRSIZE sizeof(inventory)-4  /* Size of direct area, alignment to 4 */

int              DmMain( );
inventory *      DmChItem();
inventory *      DmIsUnfixed();
int              DmStore();
void             DmRead();
unsigned short int       DmCrNew();
void             DmAddFrq();
void             DmDisConn();
void             DmConnect();
int              DmDirect();
int              DmUndirect();
void             DmObjExtend();
char *           DmFindArea();
char *           DmFind1st();
char *           DmFind2nd();
char *           DmFind3rd();
int              DmHasUnfix();

static char DISKFILE[ 63 ] = "dynamic.mem";     /* Name of working file */
static FILE *FILEHANDLE = NULL;         /* Filehandle of working file   */
static char *BEGPOOL = NULL;            /* Beginning address of main pool */
static char *BEGPOOLNAL = NULL;         /* Beginning address of main pool,
					   not aligned */
static char *ENDPOOL;                   /* Ending address of main pool  */
static inventory *INVENT;               /* Entry to the inventory       */
static unsigned short int ENT_NE_MAIN;          /* Entry to next  list of main  */
static unsigned short int ENT_NE_SECO;          /* Entry to next  list of secondary */


static unsigned short int RELEAS1 = 0;
static unsigned short int RELEAS2 = 0;
static unsigned short int RELEAS3 = 0;


int DM_ERROR = 1;       /*  Errors:
				b0 - uninitialized dynamic memory,
				b1 - lack of memory,
				b2 - wrong name of an object,
				b3 - file reading error,
				b4 - cannot find enough piece of memory,
				b5 - wrong (e.g. negative) parameter
				b6 - prohibited operation on fixed object,
				b7 - too many objects,
				b8 - wrong FROM parameter,
				b9 - cannot open secondary file,
				b10 - file writing error,
				b11 - unallowed operation on unfixed object,
				b12 - (warning) attempt too much cut the object
				      or send too much data from the object
				b13 - fixing overflow
				b14 - unfixing overflow
				b15 - wrong( e.g. negative) parameter
							*/
int DM_COMPLAIN = 0;

static long SUMFRQ;             /* Summarized frequencies */
static unsigned short int LASTFIT;      /* Last fit free area in main */
static unsigned short int OBJNBR;               /* Number of objects */

#define RESINV 8

#define It(N) (INVENT+(N)) 
#define DmMainSize(ITEM) ((long)((ITEM)->U.MAIN.OBJEND-(ITEM)->U.MAIN.OBJBEG))
/*
#define DmItemSize(ITEM) ((long)((ITEM)->STATUS & DIRECT ?                    \
			(ITEM)->STATUS & 0x0F :                               \
			(ITEM)->STATUS & SECONDARY ?                          \
			(ITEM)->U.SECOND.OBJEND - (ITEM)->U.SECOND.OBJBEG :   \
			(ITEM)->U.MAIN.OBJEND - (ITEM)->U.MAIN.OBJBEG ) )
*/
/* Note alignment to 4 */
#define DmObjAddr(ITEM) ((ITEM)->STATUS & DIRECT ? (char*)(ITEM)+4:           \
						(ITEM)->U.MAIN.OBJBEG)

/*-------------------------------------------------------- Align -----------*/
/* To avoid problems with alignment on SUN, each allocated DM object is 
   aligned to 2, what is sufficient for int, unsigned, long, etc. The
   following function makes alignement for char * pointers, i.e. it adds 1
   to the pointer if it is odd. Note that for alignment requirement, each
   structure allocated in DM should be aligned, that is, each number
   should be preceeded by even number of bytes. DM assumes that all objects
   after all operations are aligned with respect to their beginning */

/* However, bugs suggested that the alignment for long is 4-bytes boundary,
   what makes new problems */

static char *Align( CHRPNTR )
    char *CHRPNTR;
  {
/* 2-bytes boundary */
/*    return( (char *)(((unsigned)CHRPNTR + 1) & 0xFFFFFFFE) ); */
/* 4-bytes boundary */
    return( (char *)(((unsigned long)CHRPNTR + 3) & 0xFFFFFFFCL) );
  }


/*-------------------------------------------------------- DmError ---------*/
static void  DmError( CODE, NAME )
  {
    int  MESSAGE;
    int I;
    char INFOR[ 81 ];
    unsigned short int MASK = 0x0001;
    DmCheckCons();
    DM_ERROR |= CODE;
    printf( "DM_ERROR" );
    for( I = 0; I < 16; I++ )
      {
	if( DM_ERROR & MASK )
	    printf( " b%d", I );
	MASK = MASK << 1;
      }
    printf( " NAME=%u", NAME );

    switch( CODE )
      {
	case  b1:
	    MESSAGE = sDM_NO_MEMORY;
	    break;
	case  b2:
	    MESSAGE = sDM_WRONG_DM;
	    break;
	case  b3:
	    MESSAGE = sDM_WRONG_DM_FILE_READ;
	    break;
	case  b4:
	    MESSAGE = sDM_NO_MEMORY;
	    break;
	case  b6:
	    MESSAGE = sDM_PROHIBIT_OPERATION;
	    break;
	case  b7:
	    MESSAGE = sDM_TOO_MANY_DM_OBJECTS;
	    break;
	case  b9:
	    MESSAGE = sDM_WRONG_DM_FILE_OPEN;
	    break;
	case b10:
	    MESSAGE = sDM_WRONG_DM_FILE_WRITE;
	    break;
	case b15:
	    printf( "\nAlignment error\n" );
	    MESSAGE = sDM_VIOLATED_ASSERTION;
	    break;
      }
    sprintf( INFOR, FindString( MESSAGE ), NAME );
    printf( INFOR ); printf( "\n" );
    getchar();
    exit(1);
  }


/*-------------------------------------------------------- DmChItem --------*/
/* Check if object NAME is correct. Return the pointer to inventory item or
   NULL iff error. */

static inventory *   DmChItem( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
     /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
    if( (char *)ITEM < INVENT->U.MAIN.OBJEND && ITEM->STATUS && NAME )
	return( ITEM );
    DmError( b2, NAME );
    return( NULL );
  }

#define MINPOOL 2000L           /* Minimal main pool */
#define NBR_BEG_IT 100          /* Inventory size at the beginning */


/*------- External --------------------------------------- DmClose ---------*/
/* Free dynamic memory; BEGPOOLNAL - the pool; DISKFILE - name of working file */

void DmClose()
  {
				/* Delete the file */
    if( FILEHANDLE != NULL )
      { 
	fclose( FILEHANDLE );
	unlink( DISKFILE );
	FILEHANDLE = NULL;
      }
    DM_ERROR = 0;
				/* Release memory */
    if( BEGPOOLNAL != NULL )
      {
	free( BEGPOOLNAL );
	BEGPOOLNAL = NULL;
      }
  }


/*------- External --------------------------------------- DmInit ----------*/
/* Initialize dynamic memory; SIZEPOOL - the size of pool; DISKSPACE - name of
   working file, or NULL for standard/previous name. It is typically assumed 
   that DM takes all available memory, leaving some 4K for DOS */

void DmInit( SIZEPOOL, DISKSPACE )
    unsigned SIZEPOOL;
    char * DISKSPACE;
  {
    inventory *ITEM;
    DmClose();
    if( DISKSPACE != NULL )
	strcpy( DISKFILE, DISKSPACE );

    /* Get memory */
    do
      {
	BEGPOOLNAL = (char *)malloc( SIZEPOOL+3);    /* 3 added for alignment */
	if( BEGPOOLNAL == NULL )
	    SIZEPOOL -= 100;              /* If no enough memory, escape */
      }
    while( BEGPOOLNAL == NULL && SIZEPOOL > MINPOOL );

    if( BEGPOOLNAL == NULL )
	DmError( b1, 0 );
    else
      {
	long L_POOL;

	/* Alignment */
	BEGPOOL = Align( BEGPOOLNAL ); 
	ENDPOOL = BEGPOOL + SIZEPOOL;
	/* clear pool */
	for( L_POOL = SIZEPOOL; L_POOL > 0x4000; L_POOL -= 0x4000 )
	    memset( ENDPOOL-L_POOL, 0xff, 0x4000 );
	memset( ENDPOOL-L_POOL, 0xff, (int)L_POOL );
	/* Description of inventory within inventory */
	INVENT = (inventory *)BEGPOOL;
	INVENT->STATUS = EXISTING; /* existing, main, non-direct, permanent */
	INVENT->FIXING = 0;             /* unfixed */
	INVENT->U.MAIN.OBJBEG = BEGPOOL;
	/* NBR_BEG_IT cancelled objects as a beginning reserve */
	INVENT->U.MAIN.OBJEND = BEGPOOL + (NBR_BEG_IT+1) * sizeof( inventory );
	for( ITEM = It( 1 ); ITEM <= It( NBR_BEG_IT ); ITEM++ )
	    ITEM->STATUS = 0;
	INVENT->PRIOR  = 0xFFFF;
	INVENT->NEXT   = 0xFFFF;
	INVENT->RESERVE = 4 * sizeof( inventory ); /*4 items standard reserve*/
	ENT_NE_MAIN = 0;                /* Entry to next  list of main */
	ENT_NE_SECO = 0xFFFF;           /* Entry to next  list of secondary */
	LASTFIT = 0;
	SUMFRQ = 0L;
	OBJNBR = 0;
	if( FILEHANDLE == NULL )
	    FILEHANDLE = fopen( DISKFILE, "w+" );
	if( FILEHANDLE == NULL )
	    DmError( b9, 0 );
      }
  }


/*-------------------------------------------------------- DmDisConn -------*/
/* Disconnect PRIOR - NEXT connections with the object */

static void  DmDisConn( ITEM )
    inventory *ITEM;
  {
    if( ! ( ITEM->STATUS & DIRECT ) )
      {
	if( ITEM == INVENT + LASTFIT )
	    LASTFIT = ITEM->PRIOR;
	if( ITEM->NEXT != 0xFFFF )
	    It( ITEM->NEXT )->PRIOR = ITEM->PRIOR;
	if( ITEM->PRIOR != 0xFFFF )
	    It( ITEM->PRIOR )->NEXT = ITEM->NEXT;
	else
	    if( ITEM->STATUS & SECONDARY )
		ENT_NE_SECO = ITEM->NEXT;
	    else
		ENT_NE_MAIN = ITEM->NEXT;
      }
  }


/*-------------------------------------------------------- DmConnect -------*/
/* Insert an item in a proper list (main/secondary) after AFTER item and
before BEFORE item. Status of the ITEM must be set before call */

static void  DmConnect( NAME, AFTER, BEFORE )
    unsigned short int NAME, AFTER, BEFORE;
  {
    inventory *ITEM = It( NAME );
    ITEM->PRIOR = AFTER;
    ITEM->NEXT  = BEFORE;
    if( AFTER != 0xFFFF )
	It( AFTER )->NEXT = NAME;
    else
	if( ITEM->STATUS & SECONDARY )          /* If secondary list */
	    ENT_NE_SECO = NAME;
	else                                    /* If main list */
	    ENT_NE_MAIN = NAME;
    if( BEFORE != 0xFFFF )
	It( BEFORE )->PRIOR = NAME;
  }


/*-------------------------------------------------------- DmItemSize ------*/
static long  DmItemSize( ITEM )
    inventory *ITEM;
  {
    return( ITEM->STATUS & DIRECT ? (long)(ITEM->STATUS & 0x0F) :
	    ITEM->STATUS & SECONDARY ?
		ITEM->U.SECOND.OBJEND - ITEM->U.SECOND.OBJBEG :
		ITEM->U.MAIN.OBJEND - ITEM->U.MAIN.OBJBEG );
  }


/*-------------------------------------------------------- DmDirect --------*/
/* Send the object from pool to direct area. Return 0 iff unsuccessful */

static int  DmDirect( ITEM )
    inventory *ITEM;
  {
    long SIZE = DmItemSize( ITEM );
    if( ( ITEM->STATUS & SECONDARY ) || ITEM->FIXING || SIZE > DIRSIZE )
	return( 0 );
    if( ! ( ITEM->STATUS & DIRECT ) )
      {
	DmDisConn( ITEM );
/* Note alignment to 4 */
	memcpy( (char *)ITEM + 4, ITEM->U.MAIN.OBJBEG, (unsigned short int)SIZE );
	ITEM->STATUS = (unsigned short int)SIZE | ( EXISTING | DIRECT );
      }
    return( 1 );
  }


/*-------------------------------------------------------- DmStore ---------*/
/* Store the object in secondary storage. The object cannot be fixed;
   no action for direct. Return 0 iff error */

static int  DmStore( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
    unsigned short int AFTER, BEFORE;
    long BEGAREA, SIZE, OBJSIZE;
    char *STARTOBJ;
    if( ITEM->STATUS & SECONDARY || DmDirect( ITEM ) )
	return( 1 );
    SIZE = OBJSIZE = DmMainSize( ITEM );
	/* Find BEGAREA - beginning of sufficient file area */
    for( AFTER = 0xFFFF, BEFORE = ENT_NE_SECO;;)
      {
	BEGAREA = AFTER == 0xFFFF ? 0L : It( AFTER )->U.SECOND.OBJEND;
	if( BEFORE == 0xFFFF )
	    break;
	if( It( BEFORE )->U.SECOND.OBJBEG - BEGAREA >= SIZE )
	    break;
	AFTER = BEFORE;
	BEFORE = It( BEFORE )->NEXT;
      }

    STARTOBJ = DmObjAddr( ITEM );

	/* Seek on file */
    if( fseek( FILEHANDLE, BEGAREA, 0 ) == -1 )
      {
	DmError( b10, 0 );
	return( 0 );
      }
    if( fwrite( STARTOBJ, (unsigned)SIZE, 1, FILEHANDLE ) != 1 )
      {
	DmError( b10, 0 );
	return( 0 );
      }
	/* Change inventory */
    LASTFIT = ITEM->PRIOR;
    DmDisConn( ITEM );
    ITEM->STATUS = EXISTING | SECONDARY;
    ITEM->U.SECOND.OBJBEG = BEGAREA;
    ITEM->U.SECOND.OBJEND = BEGAREA + OBJSIZE;
    DmConnect( NAME, AFTER, BEFORE );
    return( 1 );
  }


/*-------------------------------------------------------- DmFind1st -------*/
/* Find Area - 1st stage: look for first fit area, beginning from LASTFIT.
Returns NULL if required area is not found */

static char *   DmFind1st( OBJSIZE, AFTER, BEFORE )
    long OBJSIZE;
    unsigned short int *AFTER, *BEFORE;
  {
    int NT;
    if( OBJSIZE <= 0 )
      {
	*AFTER = 0xFFFF;
	*BEFORE = ENT_NE_MAIN;
	return( BEGPOOL );
      }
    *AFTER = LASTFIT;
    *BEFORE = LASTFIT == 0xFFFF ? ENT_NE_MAIN : It( LASTFIT )->NEXT;
    for( NT = 1; *AFTER != LASTFIT || NT; NT = 0 )
      {
	inventory *AFTITEM = It( *AFTER );
	char *BEGA = *AFTER == 0xFFFF ? BEGPOOL : AFTITEM->U.MAIN.OBJEND;
	long FREE_AREA = 
	    ( *BEFORE == 0xFFFF ? ENDPOOL :
					It( *BEFORE )->U.MAIN.OBJBEG ) - BEGA;
	if( FREE_AREA >= OBJSIZE )
	  {
	    LASTFIT = *AFTER;
	    if( *AFTER != 0xFFFF && AFTITEM->RESERVE )
	      {
		if( FREE_AREA >= 2 * AFTITEM->RESERVE + OBJSIZE )
		    BEGA += 2 * AFTITEM->RESERVE; 
		else if( FREE_AREA >= AFTITEM->RESERVE + OBJSIZE )
		    BEGA += AFTITEM->RESERVE; 
	      }
	    return( BEGA );
	  }
	*AFTER = *BEFORE;
	*BEFORE = *BEFORE == 0xFFFF ? ENT_NE_MAIN : It( *BEFORE )->NEXT;
      }
    return( NULL );
  }


/*-------------------------------------------------------- DmFind2nd -------*/
/* Find Area - 2nd stage: look for low frequency unfixed (not-inventory) object
   and send it to secondary storage (or to direct); then repeat 1st stage.
   Return NULL if no area can be found */

static char *   DmFind2nd( OBJSIZE, AFTER, BEFORE )
    long OBJSIZE;
    unsigned short int *AFTER, *BEFORE;
  {
    unsigned short int CURRENT, RAREST;
    int RARFRQ;
    inventory *CURITEM;
    for( RAREST = 0xFFFF, RARFRQ = 256, CURRENT = ENT_NE_MAIN; 
	 CURRENT != 0xFFFF; CURRENT = CURITEM->NEXT )
      {
	CURITEM = It( CURRENT );
	if( ! CURITEM->FIXING && CURRENT )
	    if( CURITEM->FRQ < RARFRQ )
	      {
		RARFRQ = CURITEM->FRQ;
		RAREST = CURRENT;
		if( RARFRQ <= 100 )
		    break;
	      }
      }
    if( RAREST != 0xFFFF )
      {
	*AFTER  = It( RAREST )->PRIOR;
	*BEFORE = It( RAREST )->NEXT;
	if( DmStore( RAREST ) )
	  {
	    char *BEGA = *AFTER == 0xFFFF ? BEGPOOL : It(*AFTER)->U.MAIN.OBJEND;
	    if(( *BEFORE==0xFFFF ? ENDPOOL : It(*BEFORE)->U.MAIN.OBJBEG ) - BEGA
								 >= OBJSIZE )
		return( BEGA );
	  }
      }
    return( NULL );
  }


/*------- External --------------------------------------- DmCopy ----------*/
/* Copy SIZE bytes from FROM to TO */

void DmCopy( FROM, TO, SIZE )
    char *FROM, *TO;
    long SIZE;
  {
    /* bcopy( FROM, TO, (int)SIZE );  */
    memmove( TO, FROM, (int)SIZE ); 
  }


/*--------------------------------------------------------- DmFind3rd ------*/
/* Find Area - 3rd stage: sending an (unfixed) object to another place in main
   memory to obtain sufficient place. Return NULL if unsuccessful */

static char *   DmFind3rd( OBJSIZE, AFTER, BEFORE )
    long OBJSIZE;
    unsigned short int *AFTER, *BEFORE;
  {
    unsigned short int CURRENT;
    inventory *CURITEM;
    char *BEGA, *ENDA;
    long MAXOB = 20000L;
    *AFTER  = 0xFFFF;
    CURRENT = ENT_NE_MAIN;
    *BEFORE = It( CURRENT )->NEXT;
    do
      {
	CURITEM = It( CURRENT );
	if( ! CURITEM->FIXING )
	  {
	    unsigned short int AFTX, BEFX;
	    BEGA =  *AFTER == 0xFFFF ? BEGPOOL : It( *AFTER  )->U.MAIN.OBJEND;
	    ENDA = *BEFORE == 0xFFFF ? ENDPOOL : It( *BEFORE )->U.MAIN.OBJBEG;
	    if( ENDA - BEGA >= OBJSIZE )
	      {
		char *CURBEG;
		char *CUREND;
		long CURSIZE;
		char *STCURAREA;
		if( DmDirect( CURITEM ) )
		    return( BEGA );
		    /* An attempt to find free area for CURRENT object */
		if( (CURSIZE = DmMainSize( CURITEM )) < MAXOB )
		  {   
		    CURBEG = CURITEM->U.MAIN.OBJBEG;
		    CUREND = CURITEM->U.MAIN.OBJEND;
		    CURITEM->U.MAIN.OBJBEG = ENDA - OBJSIZE;
		    CURITEM->U.MAIN.OBJEND = ENDA;
		    if( ( STCURAREA = DmFind1st( CURSIZE+3, &AFTX, &BEFX ) ) 
									!= NULL)
					/* +3 because alignment */
		      {
			STCURAREA = Align( STCURAREA ); /* Alignment */
			if( AFTX != CURRENT && BEFX != CURRENT )
			  {
			    DmDisConn( CURITEM );
			    DmConnect( CURRENT, AFTX, BEFX );
			  }
			else
			    *AFTER = CURRENT;
			DmCopy( CURBEG, STCURAREA, CURSIZE );
			if( CURRENT == 0 )      /* If inventory object */
			  {
			    INVENT = (inventory *)STCURAREA;
			    CURITEM = INVENT;
			  }
			CURITEM->U.MAIN.OBJBEG = STCURAREA;
			CURITEM->U.MAIN.OBJEND = STCURAREA + CURSIZE;
			if( *AFTER != 0xFFFF)
			    BEGA = It( *AFTER )->U.MAIN.OBJEND;
			if( *AFTER == 0xFFFF )
			    BEGA = BEGPOOL;
			return( BEGA );
		      }
		    else
		      {
			MAXOB = CURSIZE;
			CURITEM->U.MAIN.OBJBEG = CURBEG;
			CURITEM->U.MAIN.OBJEND = CUREND;
		      }
		  }
	      }
	  }
	*AFTER = CURRENT;
	CURRENT = *BEFORE;
	*BEFORE = *BEFORE == 0xFFFF ? ENT_NE_MAIN : It( *BEFORE )->NEXT;
      }
    while( CURRENT != 0xFFFF );
    return( NULL );
  }


/*-------------------------------------------------------- DmFindArea ------*/
/* Find free area in main memory of required OBJSIZE. Return to AFTER the name
of an object after which the area is found, and to BEFORE the name of an object
before which the area is found. Return NULL if required area is not found. */
/* Because of alignment, the claimed area is greater on 3 */

static char *   DmFindArea( OBJSIZE, AFTER, BEFORE )
    long OBJSIZE;
    unsigned short int *AFTER, *BEFORE;
  {
    char *STARTAREA = NULL;
    if( OBJSIZE <= 0 )
      {
	*AFTER = 0xFFFF;
	*BEFORE = ENT_NE_MAIN;
	STARTAREA = BEGPOOL;
      }
    /* 1st stage: look for the first fit area, beginning from LASTFIT */
    else if( ( STARTAREA = DmFind1st( OBJSIZE, AFTER, BEFORE ) ) == NULL )
      {
	unsigned short int CURRENT;
	inventory *ITEM;
	OBJSIZE += 3; /* For alignment on SUN */
	do
	  {
	    /* 2nd stage: look for the lowest frequency unfixed (not-inventory)
	       object and send it to secondary file; then repeat 1st stage */
	    if( DM_ERROR & ( b9 | b10 ) )
		break;
	    if( ( STARTAREA = DmFind2nd( OBJSIZE, AFTER, BEFORE ) ) != NULL )
		break;
	    DM_COMPLAIN = 1;
	    if( ( STARTAREA = DmFind2nd( OBJSIZE, AFTER, BEFORE ) ) != NULL )
		break;
	    /* 3rd stage: send an (unfixed) object to another place in
	       memory to obtain sufficient place */
	    if( ( STARTAREA = DmFind3rd( OBJSIZE, AFTER, BEFORE ) ) != NULL )
		break;
	/* check if unfixed, non-direct, non-inventory, main object is found */
	    for( CURRENT = ENT_NE_MAIN; CURRENT != 0xFFFF;
						CURRENT = ITEM->NEXT )
	      {
		ITEM = It( CURRENT );
		if( ITEM->STATUS == EXISTING && !ITEM->FIXING && CURRENT )
		    break;
	      }
	  }
	while( CURRENT != 0xFFFF );
      }
    if( STARTAREA == NULL )
      {
	DmError( b4, 0 );
	return( NULL );
      }
    else
	return( Align( STARTAREA ) );
  }


/*-------------------------------------------------------- DmUndirect ------*/
/* Make the main object undirect */

static int  DmUndirect( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
    unsigned short int AFTER, BEFORE;
    char *BEGAREA;
    if( ITEM->STATUS & DIRECT ) /* If direct object */
      {
	long OBJSIZE = DmItemSize( ITEM );
	if( ( BEGAREA = DmFindArea( OBJSIZE, &AFTER, &BEFORE ) ) != NULL)
	  {
	    ITEM = It( NAME );
/* Note alignment to 4 */
	    memcpy( BEGAREA, (char *)ITEM + 4, (unsigned)OBJSIZE );
	    ITEM->STATUS = EXISTING;
	    ITEM->FIXING = ITEM->RESERVE = 0;
	    ITEM->U.MAIN.OBJBEG = BEGAREA;
	    ITEM->U.MAIN.OBJEND = BEGAREA + OBJSIZE;
	    DmConnect( NAME, AFTER, BEFORE );
	  }
	else
	    return( 0 );
      }
    return( 1 );
  }


/*-------------------------------------------------------- DmMain ----------*/
/* Send an object from secondary storage to main memory, if necessary */
/* Function returns 0 iff error */

static int  DmMain( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
    char *BEGAREA;
    long OBJSIZE;
    unsigned short int AFTER, BEFORE;

    if( ITEM->STATUS & SECONDARY )
      {
	OBJSIZE = DmItemSize( ITEM );
	if( ( BEGAREA = DmFindArea( OBJSIZE + 2 * ITEM->RESERVE,
						&AFTER, &BEFORE )) == NULL )
	    return( 0 );
	ITEM = It( NAME );
	DmDisConn( ITEM );      /* Disconnect next-prior connections */
	DmRead( ITEM, BEGAREA );
	ITEM->STATUS = EXISTING;
	ITEM->U.MAIN.OBJBEG = BEGAREA;
	ITEM->U.MAIN.OBJEND = BEGAREA + OBJSIZE;
	DmConnect( NAME, AFTER, BEFORE );
      }
    return( 1 );
  }


/*------------------------------------------------------- DmObjExtend ------*/
/* Extend the object with SIZE bytes at the top. DM_ERROR |= b4 if the object
cannot be extended */

static void  DmObjExtend( NAME, SIZE )
    unsigned short int NAME;
    long SIZE;
  {
    if( SIZE > 0 )
      {
	char *BEGAREA;
	unsigned short int AFTER, BEFORE;
	inventory *ITEM = It( NAME );
	unsigned short int NEXT;
	char *BG_NEXT;
	if( ITEM->STATUS & SECONDARY )
	  {
	    if( ! DmMain( NAME ) )
		return;
	    ITEM = It( NAME );
	  }
	if( ITEM->STATUS & DIRECT )
	  {
	    int L = ITEM->STATUS & 0x0F;
	    if( L + SIZE <= DIRSIZE )
	      {
		ITEM->STATUS += SIZE;
		return;
	      }
	    else if( L )
	      {
		if( ! DmUndirect( NAME ) )
		    return;
		ITEM = It( NAME );
	      }
	    else        /* The old size is 0 */
	      {
		if( ( BEGAREA = DmFindArea( SIZE, &AFTER, &BEFORE ) ) != NULL )
		  {
		    ITEM = It( NAME );
		    ITEM->STATUS = EXISTING;
		    ITEM->U.MAIN.OBJBEG = BEGAREA;
		    ITEM->U.MAIN.OBJEND = BEGAREA + SIZE;
		    ITEM->RESERVE = ITEM->FIXING = 0;
		    DmConnect( NAME, AFTER, BEFORE );
		    return;
		  }
	      }
	  }
	else if( ITEM->RESERVE < SIZE )
	    ITEM->RESERVE = SIZE <= 100 ? SIZE : 0;
	NEXT = ITEM->NEXT;
	BG_NEXT = NEXT == 0xFFFF ? ENDPOOL : It( NEXT )->U.MAIN.OBJBEG;
	/* Look for reserve after the object */
	if( ITEM->U.MAIN.OBJEND + SIZE <= BG_NEXT )
	    ITEM->U.MAIN.OBJEND += SIZE;
	else
	  {
	    long OBJSIZE = DmMainSize( ITEM ); 
	    /* If the next object is not inventory, is not fixed, is small,
	    makes appropriate reserve, and has frequency lower than assumed,
	    send it to secondary */
	    if( NEXT && NEXT != 0xFFFF )
	      {
		inventory *NEXT_ITEM = It( NEXT );
		if( ! NEXT_ITEM->FIXING &&
				DmMainSize( NEXT_ITEM ) < OBJSIZE + SIZE &&
				NEXT_ITEM->FRQ < 130 )
		  {
		    unsigned short int NEXT_NEXT = NEXT_ITEM->NEXT;
		    char *BG_NEXT_NEXT = NEXT_NEXT == 0xFFFF ? ENDPOOL :
						It( NEXT_NEXT )->U.MAIN.OBJBEG;
		    if( ITEM->U.MAIN.OBJEND + SIZE <= BG_NEXT_NEXT )
		      {
			DmStore( NEXT );
			ITEM->U.MAIN.OBJEND += SIZE;
			return;
		      }
		  }
	      }
	    /* Look for new space of sufficient size */ 
	    /* Set fixing to avoid sending the object to secondary file */
	    ITEM->FIXING++;
	    if( ( BEGAREA = DmFindArea( OBJSIZE + SIZE + 2 * ITEM->RESERVE,
						&AFTER, &BEFORE ) ) != NULL )
	      {
		char *OLD_BEG;
		ITEM = It( NAME );
		OLD_BEG = ITEM->U.MAIN.OBJBEG;
		ITEM->FIXING--;
		ITEM->U.MAIN.OBJBEG = BEGAREA;
		ITEM->U.MAIN.OBJEND = BEGAREA + OBJSIZE + SIZE;
		if( NAME != AFTER && NAME != BEFORE )
		  {
		    DmDisConn( ITEM );
		    DmConnect( NAME, AFTER, BEFORE );
		  }
		if( NAME == 0 ) /* If inventory object */
		    INVENT = (inventory *)BEGAREA;
		DmCopy( OLD_BEG, BEGAREA, OBJSIZE );
	      }
	    else
		ITEM->FIXING--;
	  }
      }
  }


/*-------------------------------------------------------- DmCrNew ---------*/
/* Create a new object; return 0 iff error or NAME of the object otherwise */

static unsigned short int  DmCrNew( ITEM, SIZE ) 
    inventory *ITEM;    /* Inventory item */
    unsigned SIZE;      /* Required size */
  {
    char *BEGAREA;
    unsigned short int AFTER, BEFORE, NAME;
    if( ( NAME = (unsigned short int)( ITEM - INVENT ) ) == 0xFFFF )
	/* Too many objects */
      {
	DmError( b7, 0 );
	return( 0 );
      }
    if( SIZE <= DIRSIZE )
	ITEM->STATUS = (unsigned)SIZE | ( EXISTING | DIRECT );
    else if( ( BEGAREA = DmFindArea( (long)SIZE, &AFTER, &BEFORE ) ) != NULL )
      {
	ITEM = It( NAME );
	ITEM->STATUS = EXISTING;
	ITEM->FIXING = ITEM->RESERVE = 0;
	ITEM->U.MAIN.OBJBEG = BEGAREA;
	ITEM->U.MAIN.OBJEND = BEGAREA + SIZE;
	DmConnect( NAME, AFTER, BEFORE );
	LASTFIT = NAME;
      }
    else
	return( 0 );
    ITEM->FRQ = 127;
     /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
    return( NAME );
  }


/*------- External --------------------------------------- DmCreate --------*/
/* Create an object of given SIZE and return its number */

unsigned short int DmCreate( SIZE )
    unsigned SIZE;
  {
    inventory *ITEM, *INV_END;
    unsigned short int NAME;
    OBJNBR++;
    if( RELEAS1 )
      {
	ITEM = It( RELEAS1 );
	RELEAS1 = RELEAS2;
	RELEAS2 = RELEAS3;
	RELEAS3 = 0;
	NAME = DmCrNew( ITEM, SIZE );
      }
    else
      {
	for( ITEM = It( 1 ); ITEM < (inventory *)(INVENT->U.MAIN.OBJEND);
								ITEM++ )
	    if( ! (ITEM->STATUS) )
	      {
		NAME = DmCrNew( ITEM, SIZE );
		return( NAME );
	      }
	/* Extend inventory */
	DmObjExtend( 0, 4 * sizeof( inventory ) );
	INV_END = (inventory *)( INVENT->U.MAIN.OBJEND ) - 1;
	NAME = DmCrNew( INV_END, SIZE );
	RELEAS1 = INV_END - 1 - INVENT;
	RELEAS2 = RELEAS1 - 1;
	RELEAS3 = RELEAS1 - 2;
	It(RELEAS1)->STATUS = It(RELEAS2)->STATUS = It(RELEAS3)->STATUS = 0;
      }
    return( NAME );
  }


/*-------------------------------------------------------- DmIsUnfixed -----*/
/* If the object exists and is unfixed return pointer to inventory item */

static inventory *   DmIsUnfixed( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
    if( (char *)ITEM < INVENT->U.MAIN.OBJEND && ITEM->STATUS && NAME )
	if( ! (ITEM->STATUS & DIRECT) && ITEM->FIXING ) /* If fixed */
	    DmError( b6, 0 );
	else
	    return( ITEM );
    else
	DmError( b2, NAME );
    return( NULL );
  }


/*------- External --------------------------------------- DmDelete --------*/
/* Delete an object with number determined by NAME */
/* The object cannot be fixed */

void DmDelete( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;

    if( !RELEAS1 )
	RELEAS1 = NAME;
    else if( !RELEAS2 )
	RELEAS2 = NAME;
    else
	RELEAS3 = NAME;
    OBJNBR--;
    DmDisConn( ITEM );          /* Disconnect next-prior connections */
    ITEM->STATUS = 0;           /* Cancel the inventory item */

/* The size of inventory is never shortened; about 1% of time savings
	if( RELEAS3 % 4 == 1 )
	  {
	    int NODEL = 0;
	    for( ITEM = (inventory *)( INVENT->U.MAIN.OBJEND ) - 1;
					! ITEM->STATUS; ITEM--, NODEL++ );
	    if( NODEL > NBR_BEG_IT )
	      {
		RELEAS1 = RELEAS2 = RELEAS3 = 0;
		INVENT->U.MAIN.OBJEND -=
				( NODEL - NBR_BEG_IT ) * sizeof(inventory);
	      }
	  }
*/
  }


/*------- External --------------------------------------- DmSize ----------*/
/* Return object size */
/* Function returns -1L iff error */

long DmSize( NAME )
    unsigned short int( NAME );
  {
    inventory *ITEM;
    return( ( ITEM = DmChItem( NAME ) ) == NULL ? -1L : DmItemSize( ITEM ) );
  }


/*------- External --------------------------------------- DmFrequency -----*/
/* Return object frequency */
/* Function returns -1 iff error */

int DmFrequency( NAME )
    unsigned short int( NAME );
  {
    inventory *ITEM;
    /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
    return( (ITEM = DmChItem( NAME )) == NULL ? -1 : (int)ITEM->FRQ );
  }


/*------- External --------------------------------------- DmInside --------*/
/* Return 1 iff POINTER points interior of the object, 0 otherwise or if error.
The object must be fixed */

int DmInside( POINTER, NAME )
    char *POINTER;
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
	/* No checking of existence of the object */
	/* Alignment assertion */
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
    if( ITEM->FIXING )
	return( POINTER >= ITEM->U.MAIN.OBJBEG &&
					POINTER < ITEM->U.MAIN.OBJEND );
    else
	DM_ERROR |= b11;
    return( 0 );
  }


/*-------------------------------------------------------- DmAddFrq --------*/
/* Increases frequency of the item and from time to time subtracts 1 from all
   frequencies. */

static void  DmAddFrq( ITEM )
    inventory *ITEM;
  {
	/* Assertion for alignment */
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, ITEM - INVENT );
	  }
    if(ITEM->FRQ < 255)
	ITEM->FRQ++;
    if( ++SUMFRQ > OBJNBR )
      {
	for( ITEM = It( 1 ); (char *)ITEM < INVENT->U.MAIN.OBJEND; ITEM++ )
	    if( ITEM->FRQ )
		ITEM->FRQ--;
	SUMFRQ -= OBJNBR;
      }
  }


/*------- External --------------------------------------- DmClear ---------*/
/* Make the object empty; return 0 iff error. The object cannot be fixed */

void DmClear( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    DmDisConn( ITEM );  /* Disconnect next-prior connections */
    ITEM->STATUS = EXISTING | DIRECT;
				/* existing,main,unfixed,direct,size = 0 */
    DmAddFrq( ITEM );
  }


/*------- External --------------------------------------- DmExists --------*/
/* Return 0 iff the object name is incorrect;
	 -1 iff the object is unfixed; fixing level otherwise */

int DmExists( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM = It( NAME );
    return( (char *)ITEM < INVENT->U.MAIN.OBJEND && ITEM->STATUS && NAME ?
		(ITEM->STATUS & DIRECT) || ! ITEM->FIXING ? -1 :
						(int)ITEM->FIXING : 0 );
  }


/*-------------------------------------------------------- DmRead ----------*/
/* Read an object to a place beginning from BEGAREA. No checking inside. */

static void  DmRead( ITEM, BEGAREA )
    char *BEGAREA;
    inventory *ITEM;
  {
    long OBJSIZE = DmItemSize( ITEM );
    if( fseek( FILEHANDLE, ITEM->U.SECOND.OBJBEG, 0 ) == -1 )
	DmError( b3, 0 );
    else /* if( OBJSIZE <= INT_MAX ) */
      {
	if( fread( BEGAREA, (unsigned)OBJSIZE, 1, FILEHANDLE ) != 1 )
	    DmError( b3, 0 );
      }
  }


/*------- External --------------------------------------- DmPointer --------*/
/* Check NAME, get the object from storage and return its start address. */
/* Function returns NULL iff error */

char *DmPointer( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if(( ITEM = DmChItem( NAME )) != NULL )
      {
	DmAddFrq( ITEM );
	if( ITEM->STATUS & SECONDARY && DmMain( NAME ) )
	    ITEM = It( NAME );
	return( DmObjAddr( ITEM ) );
      }
    return( NULL );
  }


/*------- External --------------------------------------- DmTopPointer ----*/
/* Check NAME, get the object from storage and return its ending address. */

char *DmTopPointer( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if(( ITEM = DmChItem( NAME )) != NULL )
      {
	DmAddFrq( ITEM );
	if( ITEM->STATUS & SECONDARY && DmMain( NAME ) )
	    ITEM = It( NAME );
	return( ITEM->STATUS & DIRECT ?
/* Note alignment to 4 */
	    ((char *)ITEM + 4 + (ITEM->STATUS & 0x0F)) : ITEM->U.MAIN.OBJEND );
      }
    return( NULL );
  }


/*------- External --------------------------------------- DmFix -----------*/
/* Check NAME, fix the object, and return its start address. */
/* Function returns NULL iff error */

char *DmFix( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if(( ITEM = DmChItem( NAME )) != NULL )
      {
	DmAddFrq( ITEM );
	if( ITEM->STATUS & ( SECONDARY | DIRECT ) )
	  {
	    if( ! ( DmMain( NAME ) && DmUndirect( NAME ) ) )
		return( NULL );
	    else
		ITEM = It( NAME );
	  }
	if( ITEM->FIXING < 255 )
	  {
	    ITEM->FIXING++;
	    return( ITEM->U.MAIN.OBJBEG );
	  }
	DM_ERROR |= b13;
      }
    return( NULL );
  }


/*------- External --------------------------------------- DmFixTop ---------*/
/* Check NAME, fix the object, and return its ending address. */
/* Function returns NULL iff error */

char *DmFixTop( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if(( ITEM = DmChItem( NAME )) != NULL )
      {
	DmAddFrq( ITEM );
	if( ITEM->STATUS & ( SECONDARY | DIRECT ) )
	  {
	    if( ! ( DmMain( NAME ) && DmUndirect( NAME ) ) )
		return( NULL );
	    else
		ITEM = It( NAME );
	  }
	if( ITEM->FIXING < 255 )
	  {
	    ITEM->FIXING++;
	    return( ITEM->U.MAIN.OBJEND );
	  }
	DM_ERROR |= b13;
      }
    return( NULL );
  }


/*------- External --------------------------------------- DmUnfix ---------*/
/* Unfix the object */

void DmUnfix( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if(( ITEM = DmChItem( NAME )) != NULL )
	if( !ITEM->FIXING-- )
	    DM_ERROR |= b14; /* unfixing overflow */
  }


/*------- External --------------------------------------- DmUnfixDelete ---*/
/* Unfix the object and then delete it */

void DmUnfixDelete( NAME )
    unsigned short int NAME;
  {
    inventory *ITEM;
    if(( ITEM = DmChItem( NAME )) != NULL )
	if( ITEM->FIXING != 1 )
	    DM_ERROR |= b14 | b6; /* unfixing overflow or illegal operation
				     on fixed object */
    if( !RELEAS1 )
	RELEAS1 = NAME;
    else if( !RELEAS2 )
	RELEAS2 = NAME;
    else
	RELEAS3 = NAME;
    OBJNBR--;
    DmDisConn( ITEM );          /* Disconnect next-prior connections */
    ITEM->STATUS = 0;           /* Cancel the inventory item */
  }


/*------- External --------------------------------------- DmPut -----------*/
/* Write data to object (bytes). Data are stored at SOURCE */
/* The object cannot be fixed */

void DmPut( NAME, SOURCE, SIZE )
    unsigned short int NAME;
    char * SOURCE;
    unsigned SIZE;
  {
    inventory *ITEM;
    char *BEGAREA;
    unsigned short int AFTER, BEFORE;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    DmAddFrq( ITEM );
    DmDisConn( ITEM );
    ITEM->STATUS = EXISTING | DIRECT;
    if( SIZE <= DIRSIZE )
      {                 /* Store directly in inventory */
	ITEM->STATUS |= SIZE;
/* Note alignment to 4 */
	memcpy( (char *)ITEM + 4, SOURCE, SIZE );
      }
    else if( ( BEGAREA = DmFindArea( (long)SIZE, &AFTER, &BEFORE )) != NULL )
      {
	memcpy( BEGAREA, SOURCE, SIZE );
	ITEM = It( NAME );
	ITEM->STATUS = EXISTING;
	ITEM->FIXING = ITEM->RESERVE = 0;
	ITEM->U.MAIN.OBJBEG = BEGAREA;
	ITEM->U.MAIN.OBJEND = BEGAREA + SIZE;
	DmConnect( NAME, AFTER, BEFORE );
      }
  }


/*------- External --------------------------------------- DmGet -----------*/
/* Get data from an object and write them into AREA */

void DmGet( NAME, AREA )
    unsigned short int NAME;
    char * AREA;
  {
    inventory *ITEM;
    if( ( ITEM = DmChItem( NAME ) ) != NULL && DmMain( NAME ) )
      {
	ITEM = It( NAME );
	DmAddFrq( ITEM );
	DmCopy( DmObjAddr( ITEM ), AREA, DmItemSize( ITEM ) );
      }
  }


/*------- External --------------------------------------- DmMove ----------*/
/* Write object NAME1 to the top of NAME2. The object NAME2 cannot be fixed */

void DmMove( NAME2, NAME1 )
    unsigned short int NAME2, NAME1;
  {
    char *B1, *B2;
    inventory *ITEM1, *ITEM2;
    long SIZE1, SIZE2;
    if( (ITEM1 = DmChItem( NAME1 )) == NULL ||
	(ITEM2 = DmIsUnfixed( NAME2 )) == NULL )
	return;
    SIZE1 = DmItemSize( ITEM1 );
    SIZE2 = DmItemSize( ITEM2 );
    B1 = DmFix( NAME1 );
    DmObjExtend( NAME2, SIZE1 );
    B2 = DmPointer( NAME2 );
    DmCopy( B1, B2 + SIZE2, SIZE1 );
    DmUnfix( NAME1 );
  }


/*------- External --------------------------------------- DmMoveTop -------*/
/* Write SIZE bytes from the top of NAME1 to the top of NAME2.
The object NAME2 cannot be fixed */

void DmMoveTop( NAME2, NAME1, SIZE )
    unsigned short int NAME2, NAME1;
    unsigned SIZE;
  {
    inventory *ITEM1, *ITEM2;
    long SIZE1, SIZE2;
    if( (ITEM1 = DmChItem( NAME1 )) == NULL ||  
	(ITEM2 = DmIsUnfixed( NAME2 )) == NULL )
	return;
    SIZE1 = DmItemSize( ITEM1 );
    SIZE2 = DmItemSize( ITEM2 );
    if( (long)SIZE > SIZE1 )
	DM_ERROR |= b12;
    else
      {
	char *B1, *B2;
	B1 = DmFixTop( NAME1 );
	DmObjExtend( NAME2, (long)SIZE );
	B2 = DmPointer( NAME2 );
	memcpy( B2 + SIZE2, B1 - SIZE, SIZE );
	DmUnfix( NAME1 );
      }
  }


/*------- External --------------------------------------- DmPush ----------*/
/* Append SIZE bytes from SOURCE to the object. The object cannot be fixed */

void DmPush( NAME, SOURCE, SIZE )
    unsigned short int NAME;
    char * SOURCE;
    unsigned SIZE;
  {
    inventory *ITEM;
    if( (ITEM = DmIsUnfixed( NAME )) != NULL )
      {
	long OBJSIZE = DmItemSize( ITEM );
	DmAddFrq( ITEM );
	DmObjExtend( NAME, (long)SIZE );
	DmCopy( SOURCE, DmObjAddr( It( NAME ) ) + OBJSIZE, SIZE );
      }
  }


/*------- External --------------------------------------- DmPushString ----*/
/* Append string from SOURCE to the object. The object cannot be fixed */

void DmPushString( NAME, SOURCE )
    unsigned short int NAME;
    char * SOURCE;
  {
    if( DmSize( NAME ) != 0L )
	DmPop( NAME, 1 );
    DmPush( NAME, SOURCE, strlen( SOURCE ) + 1 );
  }


/*------- External --------------------------------------- DmTop -----------*/
/* Send SIZE bytes from the top of the object to AREA. */

void DmTop( NAME, AREA, SIZE )
    unsigned short int NAME;
    char * AREA;
    unsigned SIZE;
  {
    inventory *ITEM;
    if( ( ITEM = DmChItem( NAME ) ) != NULL && DmMain( NAME ) )
      {
	long OBJSIZE;
	ITEM = It( NAME );
	OBJSIZE = DmItemSize( ITEM );
	DmAddFrq( ITEM );
	if( SIZE > OBJSIZE )
	    DM_ERROR |= b12;
	else
	    DmCopy( DmObjAddr( ITEM ) + OBJSIZE - SIZE, AREA, SIZE );
      }
  }


/*------- External --------------------------------------- DmPop -----------*/
/* Cut SIZE bytes from the top of an object. The object cannot be fixed */

void DmPop( NAME, SIZE )
    unsigned short int NAME;
    unsigned SIZE;
  {
    inventory *ITEM;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    DmAddFrq( ITEM );
    if( SIZE > DmItemSize( ITEM ) )
	DM_ERROR |= b12;
    else
      {
	if( ITEM->STATUS & DIRECT )  /* If direct */
	    ITEM->STATUS -= SIZE;
	else
	  {
	    if( ITEM->STATUS & SECONDARY )
	      {
		ITEM->U.SECOND.OBJEND -= SIZE;
		if( ITEM->U.SECOND.OBJEND == ITEM->U.SECOND.OBJBEG )
		  {
		    DmDisConn( ITEM );
		    ITEM->STATUS = EXISTING | DIRECT;
		  }
	      }
	    else
	      {
		ITEM->U.MAIN.OBJEND -= SIZE;
		if( ITEM->U.MAIN.OBJEND == ITEM->U.MAIN.OBJBEG ) /*  size == 0 => direct */
		  {
		    DmDisConn( ITEM );
		    ITEM->STATUS = EXISTING | DIRECT;
		  }
	      }
	  }
      }
  }


/*------- External --------------------------------------- DmTopPop --------*/
/* Send SIZE bytes from the top of the object to AREA, then cut SIZE bytes
from the top of an object. The object cannot be fixed */

void DmTopPop( NAME, AREA, SIZE )
    unsigned short int NAME;
    char * AREA;
    unsigned SIZE;
  {
    inventory *ITEM;
    if( ( ITEM = DmIsUnfixed( NAME ) ) != NULL && DmMain( NAME ) )
      {
	long OBJSIZE;
	ITEM = It( NAME );
	OBJSIZE = DmItemSize( ITEM );
	DmAddFrq( ITEM );
	if( SIZE > OBJSIZE )
	    DM_ERROR |= b12;
	else
	    DmCopy( DmObjAddr( ITEM ) + OBJSIZE - SIZE, AREA, SIZE );
	if( ITEM->STATUS & DIRECT )  /* If direct */
	    ITEM->STATUS -= SIZE;
	else
	  {
	    ITEM->U.MAIN.OBJEND -= SIZE;
	    if( ITEM->U.MAIN.OBJEND == ITEM->U.MAIN.OBJBEG ) /*  size == 0 => direct */
	      {
		DmDisConn( ITEM );
		ITEM->STATUS = EXISTING | DIRECT;
	      }
	  }
      }
  }


/*------- External --------------------------------------- DmLeave ---------*/
/* Leave SIZE bytes from FROM. The object cannot be fixed */

void DmLeave( NAME, FROM, SIZE )
    unsigned short int NAME;
    long FROM;
    unsigned SIZE;
  {
    inventory *ITEM;
    long OBJSIZE;
    int AL;     /* For alignment */
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    OBJSIZE = DmItemSize( ITEM );
    if( FROM < 0L || FROM > OBJSIZE )
	DM_ERROR |= b8;
    else
      {
	if( FROM + SIZE > OBJSIZE )
	    DM_ERROR |= b12;
	else
	  {
	    DmAddFrq( ITEM );
	    if( ITEM->STATUS & DIRECT )  /* If direct */
	      {
		ITEM->STATUS = (ITEM->STATUS & 0xF0) + SIZE;
/* Note alignment to 4 */
		DmCopy( (char *)ITEM + 4 + FROM, (char *)ITEM + 4, SIZE );
	      }
	    else if( ITEM->STATUS & SECONDARY )
	      {
		ITEM->U.SECOND.OBJBEG += FROM;
		ITEM->U.SECOND.OBJEND = ITEM->U.SECOND.OBJBEG + SIZE;
	      }
	    else /* Main memory */
	      {
		ITEM->U.MAIN.OBJBEG += FROM;
		ITEM->U.MAIN.OBJEND = ITEM->U.MAIN.OBJBEG + SIZE;
		/* Alignment problems */
		AL = FROM % 4; 
		if( AL )
		  {
		    DmCopy( ITEM->U.MAIN.OBJBEG, ITEM->U.MAIN.OBJBEG-AL, SIZE );
		    ITEM->U.MAIN.OBJBEG -= AL;
		    ITEM->U.MAIN.OBJEND -= AL;
		  }
	      }
	  }
      }
     /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
  }


/*------- External --------------------------------------- DmExtract -------*/
/* Cut SIZE bytes FROM inside of an object. The object cannot be fixed */

void DmExtract( NAME, FROM, SIZE )
    unsigned short int NAME;
    long FROM;
    unsigned SIZE;
  {
    long FROM_SIZE = FROM + SIZE;
    inventory *ITEM;
    long OBJSIZE;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    OBJSIZE = DmItemSize( ITEM );
    if( FROM < 0L || FROM > OBJSIZE )
	DM_ERROR |= b8;
    else if( FROM_SIZE > OBJSIZE )
	DM_ERROR |= b12;
    else if( FROM_SIZE == OBJSIZE )
	DmPop( NAME, SIZE );
    else
      {
	DmAddFrq( ITEM );
	if( FROM == 0L )
	  {     /* If beginning of the object */
	    if( ITEM->STATUS & DIRECT )
	      {
		unsigned NS;
		ITEM->STATUS -= SIZE;
		NS = ITEM->STATUS & 0x0F; 
		if( NS )
/* Note alignment to 4 */
		    DmCopy( (char *)ITEM + 4 + SIZE, (char *)ITEM + 4,  NS );
	      }
	    else if( ITEM->STATUS & SECONDARY )
		ITEM->U.SECOND.OBJBEG += SIZE;
	    else
	      { /* Alignment problems */
		if( SIZE % 4 )
		  {
		    DmCopy( ITEM->U.MAIN.OBJBEG + FROM_SIZE,
			ITEM->U.MAIN.OBJBEG + FROM, OBJSIZE - FROM_SIZE );
		    ITEM->U.MAIN.OBJEND -= SIZE;
		  }
		else
		    ITEM->U.MAIN.OBJBEG += SIZE;
	      }
	    if( DmItemSize( ITEM ) == 0L )      /* If size == 0, make direct */
	      {
		DmDisConn( ITEM );
		ITEM->STATUS = EXISTING | DIRECT;
	      }
	  }
	else if( DmMain( NAME ) && DmUndirect( NAME ) )
	  {
	    ITEM = It( NAME );
	    if( FROM > OBJSIZE - FROM_SIZE )
	      {
		DmCopy( ITEM->U.MAIN.OBJBEG + FROM_SIZE,
			ITEM->U.MAIN.OBJBEG + FROM, OBJSIZE - FROM_SIZE );
		ITEM->U.MAIN.OBJEND -= SIZE;
	      }
	    else
	      {
		DmCopy( ITEM->U.MAIN.OBJBEG, ITEM->U.MAIN.OBJBEG+SIZE, FROM );
		ITEM->U.MAIN.OBJBEG += SIZE;
	      } 
	  }
      }
      /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
 }


/*------- External --------------------------------------- DmDistend --------*/
/* Extend an object by SIZE bytes beginning from FROM.
DM_ERROR |= b4 if the object cannot be extended. The object cannot be fixed */

void DmDistend( NAME, FROM, SIZE )
    unsigned short int NAME;
    long FROM;
    unsigned SIZE;
  {
    inventory *ITEM;
    long OBJSIZE;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    OBJSIZE = DmItemSize( ITEM );
    if( FROM < 0L || FROM > OBJSIZE )
	DM_ERROR |= b8;
    else
      {
	DmAddFrq( ITEM );
	if( FROM == 0L &&
	    ! (SIZE % 4) &&  /* Alignment ! */
	    ! ( ITEM->STATUS & ( DIRECT | SECONDARY ) ) )
	  {
	    unsigned short int PRIOR = ITEM->PRIOR;
	    char *END_OF_PRIOR = PRIOR == 0xFFFF ? BEGPOOL :
						It( PRIOR )->U.MAIN.OBJEND; 
	    if( END_OF_PRIOR + SIZE <= ITEM->U.MAIN.OBJBEG )
	      {
		ITEM->U.MAIN.OBJBEG = ITEM->U.MAIN.OBJBEG - SIZE;
		ITEM->U.MAIN.OBJEND = ITEM->U.MAIN.OBJBEG + SIZE + OBJSIZE;
    /* Assertion for checking alignment */
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
		return;
	      }
	  }
	DmObjExtend( NAME, (long)SIZE );
	DmUndirect( NAME );
	ITEM = It( NAME );
	DmCopy( ITEM->U.MAIN.OBJBEG + FROM, ITEM->U.MAIN.OBJBEG + FROM+SIZE,
							OBJSIZE-FROM );
      }
    /* Assertion for checking alignment */
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
  }
  
/* Not used, thus may be removed */
/*------- External --------------------------------------- DmSplit ---------*/
/* Split the object into two objects, and send data beginning from FROM in
   "parent" object to "child" object */
/* Function returns 0 iff error; otherwise returns the name of child object */
/* The parent object can be fixed */
/* The child inherits all attributes from the parent, and it is located
   immediately after the parent, without garbage between */

unsigned short int DmSplit( NAME1, FROM )
    unsigned short int NAME1;
    long FROM;
  {
    inventory *ITEM1, *ITEM2;
    int FOUND;
    unsigned short int NAME2;
    if( ( ITEM1 = DmChItem( NAME1 ) ) == NULL )
	return( 0 );
    if( FROM < 0L || FROM > DmItemSize( ITEM1 ) )
      {
	DM_ERROR |= b8;
	return( 0 );
      }
    DmAddFrq( ITEM1 );
    ITEM1->STATUS &= 0xDF; /* Making permanent */
    /* Looking for new inventory item */
    FOUND = 0;
    RELEAS1 = RELEAS2 = RELEAS3 = 0;
    for( ITEM2 = It( 1 ); ITEM2 < (inventory *)INVENT->U.MAIN.OBJEND; ITEM2++ )
	if( ! (ITEM2->STATUS) )
	  {
	    FOUND = 1;
	    break;
	  }
    if( ! FOUND )
      {
	/* Extend inventory */
	DmObjExtend( 0, (long)sizeof( inventory ) );
	ITEM2 = (inventory *)( INVENT->U.MAIN.OBJEND ) - 1;
      }
    NAME2 = ITEM2 - INVENT;
    if( DmUndirect( NAME1 ) )
      {
	ITEM1 = It( NAME1 );
	ITEM2 = It( NAME2 );
	*ITEM2 = *ITEM1;
	OBJNBR++;
	ITEM1->U.MAIN.OBJEND = ITEM2->U.MAIN.OBJBEG =
					ITEM1->U.MAIN.OBJBEG + FROM;
	DmConnect( NAME2, NAME1, ITEM1->NEXT );
	return( NAME2 );
      }
    else
	return( 0 );
  } 


/*------- External --------------------------------------- DmExtend --------*/
/* Extend an object by SIZE bytes. DM_ERROR |= b4 if the object cannot be
   extended. The object cannot be fixed */

void DmExtend( NAME, SIZE )
    unsigned short int NAME;
    unsigned SIZE;
  {
    inventory *ITEM;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    DmAddFrq( ITEM );
    DmObjExtend( NAME, (long)SIZE );
	 /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
 }


/*------- External --------------------------------------- DmChangeSize -----*/
/* Change the object size to the given SIZE bytes, by cutting its top or
   extending it at its top. DM_ERROR |= b4 if the object cannot be extended.
   The object cannot be fixed */

void DmChangeSize( NAME, SIZE )
    unsigned short int NAME;
    long SIZE;
  {
    long OLDSIZE;
    inventory *ITEM;
    if( (ITEM = DmIsUnfixed( NAME )) == NULL )
	return;
    DmAddFrq( ITEM );
    OLDSIZE = DmItemSize( ITEM );
    if( SIZE < 0L )
	DM_ERROR |= b5;
    else if( OLDSIZE != SIZE )
      {
	if( OLDSIZE < SIZE )
	    DmObjExtend( NAME, SIZE - OLDSIZE );
	else if( SIZE == 0L )   /* Make direct */
	  {
	    DmDisConn( ITEM );
	    ITEM->STATUS = EXISTING | DIRECT;
	  }
	else
	  {             /* Shorten the object */
	    if( ITEM->STATUS & DIRECT )  /* If direct */
		ITEM->STATUS = SIZE;
	    else if( ITEM->STATUS & SECONDARY )
		ITEM->U.SECOND.OBJEND = ITEM->U.SECOND.OBJBEG + SIZE;
	    else
		ITEM->U.MAIN.OBJEND = ITEM->U.MAIN.OBJBEG + SIZE;
	  }
      }
     /* Assertion for checking alignment */
	ITEM = It( NAME );
    if( !( ITEM->STATUS & SECONDARY) && !(ITEM->STATUS & DIRECT) )
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    DmError( b15, NAME );
	  }
  }


/*------- External --------------------------------------- DmStatistics ----*/
/* Get basic statistical data concerning DM objects */

void DmStatistics( OBJNBR, MAINNBR, FIXNBR, TOTSIZE, MAINSIZE, MAXGAR, MAXFIX,
							    AREASIZE, INVSIZE )
    unsigned *OBJNBR;
    unsigned *MAINNBR;
    unsigned *FIXNBR;
    long *TOTSIZE;
    long *MAINSIZE;
    long *MAXGAR;
    long *MAXFIX;
    long *AREASIZE;
    unsigned *INVSIZE;
  {
    inventory *ITEM;
    unsigned IOBJNBR  = 0;
    unsigned IMAINNBR = 0;
    unsigned IFIXNBR  = 0;
    long ITOTSIZE  = 0L;
    long IMAINSIZE = 0L;

    for( ITEM = It(0); (char *)ITEM < (char *)(INVENT->U.MAIN.OBJEND); ITEM++ )
	if( ITEM->STATUS )
	  {
	    long SIZE = DmItemSize( ITEM );
	    IOBJNBR++;
	    if( ! (ITEM->STATUS & DIRECT) )
		ITOTSIZE += SIZE;
	    if( ! (ITEM->STATUS & SECONDARY) )
	      {
		IMAINNBR++;
		if( ! (ITEM->STATUS & DIRECT) )
		    IMAINSIZE += SIZE;
	      }
	    if( !( ITEM->STATUS & (DIRECT | SECONDARY) ) && ITEM->FIXING )
		IFIXNBR++;
	  }
    if( OBJNBR != NULL )
	*OBJNBR = IOBJNBR;
    if( MAINNBR != NULL )
	*MAINNBR = IMAINNBR;
    if( FIXNBR != NULL )
	*FIXNBR = IFIXNBR;
    if( TOTSIZE != NULL )
	*TOTSIZE = ITOTSIZE;
    if( MAINSIZE != NULL )
	*MAINSIZE = IMAINSIZE;
    if( AREASIZE != NULL )
	*AREASIZE = (long)( ENDPOOL - BEGPOOL );
    if( INVSIZE != NULL )
	*INVSIZE = (unsigned)( ( INVENT->U.MAIN.OBJEND -
			INVENT->U.MAIN.OBJBEG ) / sizeof( inventory ) );

    if( MAXGAR != NULL )
      {
	unsigned short int AFTER, BEFORE;
	*MAXGAR = 0L;
	for( AFTER = 0xFFFF, BEFORE = ENT_NE_MAIN; ;
	     AFTER = BEFORE, BEFORE = It( BEFORE )->NEXT )
	  {
	    long GARSIZE = ( BEFORE==0xFFFF ? ENDPOOL : It( BEFORE )->U.MAIN.OBJBEG )
			- (AFTER == 0xFFFF ? BEGPOOL : It( AFTER )->U.MAIN.OBJEND );
	    if( MAXGAR != NULL )
		if( GARSIZE > *MAXGAR )
		    *MAXGAR = GARSIZE;
	    if( BEFORE == 0xFFFF )
		break;
	  }
      }

    if( MAXFIX != NULL )
      {
	long SMAXF, INVSIZE;
	unsigned short int BEFORE, FIXAFTER, FIXBEFORE;
	*MAXFIX = SMAXF = 0L;
	FIXAFTER = FIXBEFORE = 0xFFFF;
	for( BEFORE = ENT_NE_MAIN; ; BEFORE = It( BEFORE )->NEXT )
	  {
	    if( BEFORE == 0xFFFF || It( BEFORE )->FIXING )
	      {
		long GARSIZE;
		FIXAFTER = FIXBEFORE;
		FIXBEFORE = BEFORE;
		GARSIZE = (FIXBEFORE==0xFFFF ? ENDPOOL: It(FIXBEFORE)->U.MAIN.OBJBEG)
			 - (FIXAFTER==0xFFFF ? BEGPOOL: It(FIXAFTER)->U.MAIN.OBJEND);
		if( GARSIZE > *MAXFIX )
		  {
		    SMAXF = *MAXFIX;
		    *MAXFIX = GARSIZE;
		  }
		else if( GARSIZE > SMAXF )
		    SMAXF = GARSIZE;
	      }
	    if( BEFORE == 0xFFFF )
		break;
	  }
	INVSIZE = INVENT->U.MAIN.OBJEND - INVENT->U.MAIN.OBJBEG;
	if( SMAXF < INVSIZE )
	    if( *MAXFIX - INVSIZE > SMAXF )
		*MAXFIX = *MAXFIX - INVSIZE;
	    else
		*MAXFIX = SMAXF;
      }
  }


/*-------------------------------------------------------- DmCheckCons -----*/
/* Testing function: check consistency of DM */

int DmCheckCons()
  {
    int CONS = 1;
    unsigned short int AFTER, BEFORE;
    inventory *ITEM;
    /* Check consistency of area */
    if( BEGPOOL >= ENDPOOL )
      {
	CONS = 0;
printf( "Inconsistency of DM\n" );
/*
	printf( FindString( sDM_INCONSISTENCY ), BEGPOOL, ENDPOOL );
*/
      }

    /* Check consistency of the chain of secondary objects */
    for( AFTER = 0xFFFF, BEFORE = ENT_NE_SECO; BEFORE != 0xFFFF;)
      {
	long BEGAREA = AFTER==0xFFFF ? 0L : (long)It( AFTER )->U.MAIN.OBJEND;
	long ENDAREA = (long)It( BEFORE )->U.MAIN.OBJBEG;
	if( BEGAREA > ENDAREA )
	  {
printf( "Inconsistency of DM secondary storage\n" );
/*
	    printf( FindString( sDM_INCONSIST_SECONDARY ),
			AFTER, BEGAREA, BEFORE, ENDAREA );
*/
	    return( 0 );
	  }
	AFTER = BEFORE;
	BEFORE = It( BEFORE )->NEXT;
      }
    /* Check consistency of the chain of main objects */
    for( AFTER = 0xFFFF, BEFORE = ENT_NE_MAIN; BEFORE != 0xFFFF;)
      {
	char *BEGAREA = AFTER==0xFFFF ? BEGPOOL : It( AFTER )->U.MAIN.OBJEND;
	char *ENDAREA = It( BEFORE )->U.MAIN.OBJBEG;
     /* Assertion for checking alignment */
	ITEM = It( BEFORE );
	if( Align( ITEM->U.MAIN.OBJBEG ) != ITEM->U.MAIN.OBJBEG )
	  {
	    printf( "Alignment inconsistency: object %u\n", BEFORE ); 
	  }

	if( BEGAREA > ENDAREA )
	  {
printf( "Inconsistency of DM main memory\n" );

/*
	    printf( FindString( sDM_INCONSIST_MAIN ),
			AFTER, BEGAREA, BEFORE, ENDAREA );
*/
	    return( 0 );
	  }
	AFTER = BEFORE;
	BEFORE = It( BEFORE )->NEXT;
      }
    /* Check consistency of the inventory */
    for( ITEM = It( 0 ); ITEM < (inventory *)(INVENT->U.MAIN.OBJEND); ITEM++ )
	if( ITEM->STATUS && ! ( ITEM->STATUS & DIRECT ) )
	/* If not cancelled and not direct */
	  {
	    if( ITEM->STATUS & SECONDARY )
	      {
		for( AFTER = (unsigned short int)(ITEM-INVENT), BEFORE = ITEM->NEXT;
							BEFORE != 0xFFFF; )
		  {
		    long BEGAREA = AFTER==0xFFFF ? 0L :
					It( AFTER )->U.SECOND.OBJEND;
		    long ENDAREA = It( BEFORE )->U.SECOND.OBJBEG;
		    if( BEGAREA > ENDAREA )
		      {
printf( "Inconsistency of DM secondary inventory\n" );
/*
			printf( FindString( sDM_INVENTORY_SECONDARY ),
					AFTER, BEGAREA, BEFORE, ENDAREA );
*/
			return( 0 );
		      }
		    AFTER = BEFORE;
		    BEFORE = It( BEFORE )->NEXT;
		  }
	      }
	    else
	      {
		for( AFTER = (unsigned short int)(ITEM-INVENT), BEFORE = ITEM->NEXT;
							BEFORE != 0xFFFF; )
		  {
		    char *BEGAREA = AFTER==0xFFFF ? BEGPOOL :
						It( AFTER )->U.MAIN.OBJEND;
		    char *ENDAREA = It( BEFORE )->U.MAIN.OBJBEG;
		    if( BEGAREA > ENDAREA )
		      {
printf( "Inconsistency of DM main inventory\n" );
/*
			printf( FindString( sDM_INVENTORY_MAIN ),
					AFTER, BEGAREA, BEFORE, ENDAREA );
*/
			return( 0 );
		      }
		    AFTER = BEFORE;
		    BEFORE = It( BEFORE )->NEXT;
		  }
	      }
	  }
    return( CONS );
  }


/*--------------------------------------------------------------------------*/
#undef MINPOOL
#undef DIRSIZE
#undef EXISTING
#undef SECONDARY
#undef DIRECT

/* ======================================================================== */
